

#include "AgentList.h"

#include "AgentTypeDefs.h"
#include "../misc/Memory.h"

static AgentIdNodeType *
CreateNode (AgentIdType AgentId)
{
  AgentIdNodeType *new_node;

  new_node = Allocate (sizeof (struct _AgentIdNodeType));
  new_node->PtrPrev = new_node->PtrNext = new_node;
  new_node->AgentId = AgentId;
  return new_node;
}

static void 
LinkNodeAfter (AgentIdNodeType * Node, AgentIdNodeType * NewNode)
{
  if (Node)
    {
      Node->PtrNext->PtrPrev = NewNode;
      NewNode->PtrNext = Node->PtrNext;
      Node->PtrNext = NewNode;
      NewNode->PtrPrev = Node;
    }
  else
    NewNode->PtrPrev = NewNode->PtrNext = NewNode;
}

static void 
LinkNodeBefore (AgentIdNodeType * Node, AgentIdNodeType * NewNode)
{
  if (Node)
    {
      Node->PtrPrev->PtrNext = NewNode;
      NewNode->PtrPrev = Node->PtrPrev;
      Node->PtrPrev = NewNode;
      NewNode->PtrNext = Node;
    }
  else
    NewNode->PtrPrev = NewNode->PtrNext = NewNode;
}

static void 
UnlinkNode (AgentIdNodeType * Node)
{
  if ((Node->PtrPrev != Node) && (Node->PtrNext != Node))
    {
      Node->PtrNext->PtrPrev = Node->PtrPrev;
      Node->PtrPrev->PtrNext = Node->PtrNext;
      Node->PtrPrev = Node->PtrNext = Node;
    }
}


/*****************************************************************************/

inline AgentIdNodeType *
AgentList_Head (const AgentIdListType * List)
{
  return List ? List->PtrHead : NULL;
}

inline AgentIdNodeType *
AgentList_Tail (const AgentIdListType * List)
{
  if (List)
    return List->PtrHead ? List->PtrHead->PtrPrev : NULL;
  else
    return NULL;
}

inline AgentIdNodeType *
AgentList_Pred (const AgentIdListType * List,
		const AgentIdNodeType * Node)
{
  if (Node)
    return Node == AgentList_Head (List) ? NULL : Node->PtrPrev;
  else
    return AgentList_Tail (List);
}

inline AgentIdNodeType *
AgentList_Succ (const AgentIdListType * List,
		const AgentIdNodeType * Node)
{
  if (Node)
    return Node == AgentList_Tail (List) ? NULL : Node->PtrNext;
  else
    return AgentList_Head (List);
}

AgentIdNodeType *
AgentList_NodeAt (const AgentIdListType * List, int index)
{
  AgentIdNodeType *node = List->PtrHead;

  while (index-- && node)
    node = AgentList_Succ (List, node);

  return node;
}

inline AgentIdType 
AgentList_IdAt (const AgentIdListType * List, int index)
{
  AgentIdNodeType *node;

  if ((node = AgentList_NodeAt (List, index)) != NULL)
    return node->AgentId;
  else
    return 0;
}

int 
AgentList_IndexOf (const AgentIdListType * List,
		   const AgentIdNodeType * Node)
{
  AgentIdNodeType *node = List->PtrHead;
  int index = 0;

  while (node)
    {
      if (node == Node)
	break;
      index++;
      node = AgentList_Succ (List, node);
    }

  return node ? index : -1;
}

AgentIdNodeType *
AgentList_FindAgentId (const AgentIdListType * List,
		       AgentIdType AgentId)
{
  AgentIdNodeType *node = List->PtrHead;

  while (node)
    {
      if (node->AgentId == AgentId)
	return node;
      node = AgentList_Succ (List, node);
    }
  return NULL;
}

int 
AgentList_IndexOfAgentId (const AgentIdListType * List,
			  AgentIdType AgentId)
{
  AgentIdNodeType *node = List->PtrHead;
  int index = 0;

  while (node)
    {
      if (node->AgentId == AgentId)
	return index;

      node = AgentList_Succ (List, node);
      index++;
    }
  return -1;
}

bool 
AgentList_IsMember (const AgentIdListType * List, AgentIdType AgentId)
{
  return AgentList_FindAgentId (List, AgentId) != NULL;
}



/*****************************************************************************/

void 
AgentList_InsertBefore (AgentIdListType * List, AgentIdNodeType * Node, AgentIdType AgentId)
{
  AgentIdNodeType *new_node = CreateNode (AgentId);

  LinkNodeBefore (Node, new_node);
  if (Node == List->PtrHead)
    List->PtrHead = new_node;
  List->Length++;
}

void 
AgentList_InsertAfter (AgentIdListType * List, AgentIdNodeType * Node, AgentIdType AgentId)
{
  AgentIdNodeType *new_node = CreateNode (AgentId);

  LinkNodeAfter (Node, new_node);
  if (List->PtrHead == NULL)
    List->PtrHead = new_node;
  List->Length++;
}

void 
AgentList_Prepend (AgentIdListType * List, AgentIdType AgentId)
{
  AgentList_InsertBefore (List, AgentList_Head (List), AgentId);
}

void 
AgentList_Append (AgentIdListType * List, AgentIdType AgentId)
{
  AgentList_InsertAfter (List, AgentList_Tail (List), AgentId);
}


void 
AgentList_RemoveNode (AgentIdListType * List, AgentIdNodeType * Node)
{
  if (Node == List->PtrHead)
    List->PtrHead = AgentList_Succ (List, Node);

  UnlinkNode (Node);
  List->Length--;
}

void 
AgentList_Remove (AgentIdListType * List, AgentIdType AgentId)
{
  AgentIdNodeType *node;

  if ((node = AgentList_FindAgentId (List, AgentId)) != NULL)
    {
      AgentList_RemoveNode (List, node);
      Free (node);
    }
}

AgentIdType 
AgentList_RemoveFront (AgentIdListType * List)
{
  if (List->PtrHead)
    {
      AgentIdType agentId = List->PtrHead->AgentId;

      AgentList_RemoveNode (List, List->PtrHead);
      return agentId;
    }
  else
    return 0;
}

AgentIdType 
AgentList_RemoveRear (AgentIdListType * List)
{
  if (List->PtrHead)
    {
      AgentIdType agentId = AgentList_Tail (List)->AgentId;

      AgentList_RemoveNode (List, List->PtrHead);
      return agentId;
    }
  else
    return 0;
}


/*****************************************************************************/

IterStateType 
AgentList_InitIteration (const AgentIdListType * List)
{
  return List->PtrHead;
}

AgentIdType 
AgentList_Next (const AgentIdListType * List,
		IterStateType * PtrState)
{
  AgentIdType agentId = 0;

  if (*PtrState)
    {
      agentId = (*PtrState)->AgentId;
      *PtrState = AgentList_Succ (List, *PtrState);
    }
  return agentId;
}

AgentIdType 
AgentList_Prev (const AgentIdListType * List,
		IterStateType * PtrState)
{
  AgentIdType agentId = 0;

  if (*PtrState)
    {
      agentId = (*PtrState)->AgentId;
      *PtrState = AgentList_Pred (List, *PtrState);
    }
  return agentId;
}


/*****************************************************************************/

AgentIdListType *
AgentList_CreateList ()
{
  AgentIdListType *new_list;

  new_list = Allocate (sizeof (AgentIdListType));
  new_list->PtrHead = NULL;
  new_list->Length = 0;
  return new_list;
}

AgentIdListType *
AgentList_CopyList (AgentIdListType * PtrDest,
		    const AgentIdListType * PtrSrc)
{
  AgentIdNodeType *src_node;

  if (PtrSrc)
    {
      src_node = PtrSrc->PtrHead;
      if (!PtrDest)
	PtrDest = AgentList_CreateList ();

      while (src_node)
	{
	  AgentList_Append (PtrDest, src_node->AgentId);
	  src_node = AgentList_Succ (PtrSrc, src_node);
	}
      return PtrDest;
    }
  else
    return NULL;
}

void 
AgentList_EmptyList (AgentIdListType * List)
{
  while (List->PtrHead)
    {
      AgentIdNodeType *node = List->PtrHead;

      AgentList_RemoveFront (List);
      Free (node);
    }
  List->PtrHead = NULL;
}

void 
AgentList_DestroyList (AgentIdListType * List)
{
  AgentList_EmptyList (List);
  Free (List);
}


inline bool 
AgentList_IsEmpty (const AgentIdListType * List)
{
  return List->PtrHead == NULL;
}

bool 
AgentList_IsAtom (const AgentIdListType * List)
{
  if (List->PtrHead)
    return (List->PtrHead->PtrPrev == List->PtrHead) &&
      (List->PtrHead->PtrNext == List->PtrHead);
  return 0;
}

int 
AgentList_Count (const AgentIdListType * List)
{
  int c = 0;
  AgentIdNodeType *node = AgentList_Head (List);

  while (node)
    {
      c++;
      node = AgentList_Succ (List, node);
    }
  return c;
}

inline int 
AgentList_Length (const AgentIdListType * List)
{
  return List->Length;
}
