
#include "AgentTypeDefs.h"
#include "Agent.h"
#include "../misc/Messages.h"
#include "../misc/Memory.h"
#include "AgentList.h"
#include "../dictionary/Dictionary.h"

static int DL;

typedef struct _AgentSet
  {
    DictType *Map;
    AgentIdListType *List;

    AgentIdType CurrentId;
  }
AgentSet;

inline AgentType *AgentSets_GetAgent (const AgentSet * PtrSet,
				      AgentIdType AgentId);

/*-local-functions-*/

static inline AgentIdType 
GetUnusedId (AgentSet * PtrSet)
{
  return ++PtrSet->CurrentId;
}

static inline void 
UseId (AgentSet * PtrSet, AgentIdType AgentId)
{
  if (AgentId > PtrSet->CurrentId)
    PtrSet->CurrentId = AgentId;
}

static inline bool 
IsIdValid (const AgentSet * PtrSet, AgentIdType AgentId)
{
  if (AgentId > PtrSet->CurrentId)
    return TRUE;
  else
    return AgentSets_GetAgent (PtrSet, AgentId) == NULL;
}


void 
AgentSets_Check (const AgentSet * PtrSet)
{
  IterStateType state;
  AgentIdType agent;

  state = AgentList_InitIteration (PtrSet->List);
  while ((agent = AgentList_Next (PtrSet->List, &state)) != 0)
    if (!Dictionary_Find (PtrSet->Map, agent))
      Warning ("[AgentSets]Check",
	       "set %p: agent %d in list but not in dict\n", PtrSet, agent);
}

inline AgentType *
AgentSets_GetAgent (const AgentSet * PtrSet,
		    AgentIdType AgentId)
{
  return Dictionary_Find (PtrSet->Map, (DictKeyType) AgentId);
}

inline unsigned 
AgentSets_GetCount (const AgentSet * PtrSet)
{
  return AgentList_Count (PtrSet->List);
}

AgentIdType 
AgentSets_InsertAgent (AgentSet * PtrSet, AgentIdType AgentId,
		       const AgentType * PtrAgent)
{
  if (AgentId == 0)
    AgentId = GetUnusedId (PtrSet);

  if ((AgentId != 0) && IsIdValid (PtrSet, AgentId))
    {
      UseId (PtrSet, AgentId);
      if (Dictionary_InsertItem (PtrSet->Map, PtrAgent, AgentId))
	{
	  AgentList_Append (PtrSet->List, AgentId);
	  return AgentId;
	}
    }

  Warning ("AgentSets_InsertAgent",
	   "couldn't insert agent %p with id %d into set %p\n",
	   PtrAgent, AgentId, PtrSet);

  return 0;
}

AgentType *
AgentSets_RemoveAgent (AgentSet * PtrSet, AgentIdType AgentId)
{
  AgentType *agent;

  agent = AgentSets_GetAgent (PtrSet, AgentId);
  AgentList_Remove (PtrSet->List, AgentId);
  Dictionary_Delete (PtrSet->Map, AgentId);
  return agent;
}


inline IterStateType 
AgentSets_InitAgentIteration (const AgentSet * PtrSet)
{
  return AgentList_InitIteration (PtrSet->List);
}

inline AgentIdType 
AgentSets_GetNextAgent (const AgentSet * PtrSet,
			IterStateType * PtrState)
{
/*      Check(PtrSet); */
  return AgentList_Next (PtrSet->List, PtrState);
}


void 
AgentSets_ForEach (AgentSet * PtrSet,
		   void (*todo) (AgentSet *, AgentIdType, void *),
		   void *data)
{
  IterStateType state;
  AgentIdType agentId, nextAgentId;

  state = AgentSets_InitAgentIteration (PtrSet);
  agentId = AgentSets_GetNextAgent (PtrSet, &state);
  while (agentId != 0)
    {
      nextAgentId = AgentSets_GetNextAgent (PtrSet, &state);
      (*todo) (PtrSet, agentId, data);
      agentId = nextAgentId;
    }
}

void 
AgentSets_ForEachAgent (AgentSet * PtrSet,
			void (*todo) (AgentType *, void *), void *data)
{
  IterStateType state;
  AgentIdType agentId, nextAgentId;

  state = AgentSets_InitAgentIteration (PtrSet);
  agentId = AgentSets_GetNextAgent (PtrSet, &state);
  while (agentId != 0)
    {
      nextAgentId = AgentSets_GetNextAgent (PtrSet, &state);
      (*todo) (AgentSets_GetAgent (PtrSet, agentId), data);
      agentId = nextAgentId;
    }
}


AgentSet *
AgentSets_CreateAgentSet ()
{
  AgentSet *set;

  DebugLevel (DL, "Create AgentSet\n");

  set = Allocate (sizeof (struct _AgentSet));
  set->CurrentId = 0;
  set->Map = Dictionary_Create ();
  set->List = AgentList_CreateList ();
  return set;
}

void 
AgentSets_DestroyAgentSet (AgentSet * PtrSet)
{
  IterStateType iter;
  AgentIdType agentId;

  iter = AgentList_InitIteration (PtrSet->List);
  agentId = AgentList_Next (PtrSet->List, &iter);
  while (agentId)
    {
      AgentIdType next_agent;

      next_agent = AgentList_Next (PtrSet->List, &iter);
      AgentSets_RemoveAgent (PtrSet, agentId);
      agentId = next_agent;
    }

  AgentList_DestroyList (PtrSet->List);
  Dictionary_Destroy (PtrSet->Map);

  Free (PtrSet);
}
