/*-module-*****************************************************************
 *
 * Modul     : Dictionary (Datenmodul)
 *
 * Beschreib.:
 *
 * Aender.   :
 *
 * 20.5.1996, erik: NEUIMPLENTIERUNG. Verwaltet die Elemente nicht mehr wie 
 *	 in der vorigen Version mit Hilfe einer Hashtable sondern mit direkt
 *	 addressierbaren Arrays. Der Geschwindigkeitsvorteil ist enorm (Faktor 10),
 *	 der Speicherbedarf nicht hoeher. Letzteres klappt natuerlich nur, weil
 *	 eine besondere Eigenschaft der Schluessel vorausgesetzt wird: Die 
 *	 Differenz zwischen dem groessten und kleinsten Schluessel ist relativ
 *   konstant und eher klein. (Auf jeden Fall kleiner als einige Millionen!)
 *
 *--**********************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"


#define PAGE_SIZE 	128
#define NUM_PAGES	8000	/* Damit das Dictionary kleiner als 64k bleibt.
				   (Wichtig fuer PCs??) */

typedef struct __chunk
  {
    int used;
    void **entries;
  }
PageType;


typedef struct __dict
  {
    PageType pages[NUM_PAGES];
    int used;
  }
DictType;


typedef int DictKeyType;


/*-ftn-********************************************************************
 *
 * Dictionary_Create
 *
 * Erzeugt ein leeres Dictionary
 *
 * #fertig
 *
 *************************************************************************/

DictType *
Dictionary_Create (void)
{
  static bool printedWarning = FALSE;

  if (!printedWarning)
    {
      fprintf (stderr, "WARNING: Using an experimental implementation of "
	       "dictionary. It will silently fail if difference between the "
	       "largest and smallest id is greater than %d. (This "
	"shouldn't happen too often, cf. MaxAge)\n", NUM_PAGES * PAGE_SIZE);
      printedWarning = TRUE;
    }

  return calloc (1, sizeof (DictType));
}



/*-ftn-********************************************************************
 *
 * Dictionary_Destroy
 *
 * Loescht alle Elemente aus dem Dictionary und gibt den fuer interne
 * Datenstrukturen verbrauchten Speicher frei.
 *
 * #fertig
 *
 *************************************************************************/

bool 
Dictionary_Destroy (DictType * dict)
{
  int pageno;

  for (pageno = 0; pageno < NUM_PAGES; pageno++)
    if (dict->pages[pageno].used > 0)
      {
	free (dict->pages[pageno].entries);
	dict->used--;
      }

  if (dict->used != 0)
    {
      Error ("Dictionary_Destroy", "Internal inconsistency.\n");
    }

  free (dict);

  return TRUE;
}


/*-ftn-********************************************************************
 *
 * Dictionary_ActivePages
 *
 * Liefert die Anzahl der Speicherseiten, die das Dictionary zur Zeit
 * benoetigt, um alle Ids zu verwalten.
 *
 * #fertig
 *
 *************************************************************************/

int 
Dictionary_ActivePages (DictType * dict)
{
  return dict->used;
}


/*-ftn-********************************************************************
 *
 * Dictionary_InsertItem
 *
 * Fuegt das Element 'item' unter dem Schluessel 'id' in das Dictionary
 * ein. Die Werte werden nicht kopiert.
 *
 * #fertig
 *
 *************************************************************************/

bool 
Dictionary_InsertItem (DictType * dict, void *item, DictKeyType id)
{
  int pageno, elementno;
  PageType *page;

  pageno = (id / PAGE_SIZE) % NUM_PAGES;
  elementno = id % PAGE_SIZE;
  page = &(dict->pages[pageno]);

  if (page->used == 0)
    {
      page->entries = calloc (PAGE_SIZE, sizeof (void *));
      dict->used++;
      if (dict->used > NUM_PAGES - 10)
	Warning ("Dictionary", "Dictionary is getting too full.\n");
    }

  if (page->entries[elementno] != NULL)
    {
      Error ("Dictionary_InsertItem", "Insertion of element %d failed.\n", id);
      return FALSE;
    }

  page->entries[elementno] = item;
  page->used++;

  return TRUE;
}



/*-ftn-********************************************************************
 *
 * Dictionary_Delete
 *
 * Loescht das Element das unter dem Schluessel 'id' abgelegt ist.
 *
 * #fertig
 *
 *************************************************************************/

bool 
Dictionary_Delete (DictType * dict, DictKeyType id)
{
  int pageno, elementno;
  PageType *page;

  pageno = (id / PAGE_SIZE) % NUM_PAGES;
  elementno = id % PAGE_SIZE;
  page = &(dict->pages[pageno]);

  if ((page->used == 0) || (page->entries[elementno] == NULL))
    {
      Error ("Dictionary_Delete", "Deletion failed. No item with id %d\n", id);
      return FALSE;
    }

  page->entries[elementno] = NULL;
  page->used--;

  if (page->used == 0)
    {
      free (page->entries);
      page->entries = NULL;
      dict->used--;
    }

  return TRUE;
}



/*-ftn-********************************************************************
 *
 * Dictionary_Find
 *
 * Liefert das Element das unter dem Schluessel 'id' abgelegt ist.
 *
 * #fertig
 *
 *************************************************************************/

void *
Dictionary_Find (const DictType * dict, DictKeyType id)
{
  int pageno, elementno;
  PageType *page;

  pageno = (id / PAGE_SIZE) % NUM_PAGES;
  elementno = id % PAGE_SIZE;
  page = &(dict->pages[pageno]);

  if (page->used == 0)
    return NULL;
  return page->entries[elementno];
}







/*

   Und hier das Ergebnis der Bemuehungen:

   sultan> time failure1 -s 1000 hungry.wf
   MaxTimeStep: 1000
   Creating Channel end.dump 1
   Using seed 100

   time        1, agents    54,  ##########
   time      101, agents    54,  ##########
   time      201, agents    54,  ##########
   time      301, agents    81,  ##########
   time      401, agents    66,  ##########
   time      501, agents    75,  ##########
   time      601, agents    66,  ##########
   time      701, agents    73,  ##########
   time      801, agents    61,  ##########
   time      901, agents    67,  ##########Simulation end.

   409.326u 3.249s 7:17.67 94.2% 0+0k 5+19io 0pf+0w
   sultan> cp end.dump end1.dump
   sultan> time failure2 -s 1000 hungry.wf
   MaxTimeStep: 1000
   Creating Channel end.dump 1
   WARNING: Using experimental implementation of dictionary.
   It will silently fail if largest id - smallest id > 67108864!
   WARNING: Using experimental implementation of dictionary.
   It will silently fail if largest id - smallest id > 67108864!
   Using seed 100

   time        1, agents    54,  ##########
   time      101, agents    54,  ##########
   time      201, agents    54,  ##########
   time      301, agents    81,  ##########
   time      401, agents    66,  ##########
   time      501, agents    75,  ##########
   time      601, agents    66,  ##########
   time      701, agents    73,  ##########
   time      801, agents    61,  ##########
   time      901, agents    67,  ##########Simulation end.

   33.103u 0.968s 0:37.62 90.5% 0+0k 5+24io 0pf+0w
   sultan> diff end.dump end1.dump
   sultan> 

 */
