
#include "DumpPrivate.h"
#include "Dump.h"
#include "../misc/DynaString.h"
#include "../misc/StringList.h"
#include "../misc/Messages.h"
#include "../misc/Memory.h"
#include "../misc/Resources.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <assert.h>
#include <time.h>
#include <string.h>
#include <varargs.h>
#include <errno.h>

typedef enum
  {
    Startup = 0, Running, End
  }
SimState;

static SimState simState = Startup;

static struct timeval StartUserTime, LastUserTime, CurrentUserTime;
static unsigned StartMemory, LastMemory, CurrentMemory;

static char StartTime[32], EndTime[32];
/*static struct rusage  StartResources, LastResources, CurrentResources; */

static struct timeval hertzToTimeval (int hertz);
static struct timeval subtract_times (struct timeval tv1, struct timeval tv2);
static void fprintt (FILE * stream, struct timeval timeval);
static void sprintt (char *buffer, struct timeval timeval);

static StringListType *parameterKeys = NULL, *parameterValues = NULL;

DumpChannel Default_Channel =
{
  NULL,

  "Default", NULL, IO_FileChannel,
  {'w', '\0'}, FALSE,
  NULL,
  {
    DISPLAY_NEVER,
    DISPLAY_NEVER, DISPLAY_NEVER, DISPLAY_NEVER,
    DISPLAY_NEVER, DISPLAY_NEVER, DISPLAY_NEVER, DISPLAY_NEVER,
    DISPLAY_NEVER,
    DISPLAY_NEVER, DISPLAY_NEVER, DISPLAY_NEVER, DISPLAY_NEVER,
    DISPLAY_NEVER, DISPLAY_NEVER, DISPLAY_NEVER,
    DISPLAY_NEVER, DISPLAY_NEVER, DISPLAY_NEVER,
    DISPLAY_NEVER
  },
  NULL
};

DumpChannel *Channels = NULL;

enum
  {
    TotalDumpAtEnd,		/* ist die zugehoerige Variable fuer einen Kanal gesetzt, 
				   wird beim Simulationsende ein Totaldump auf diesem 
				   Kanal ausgegeben. 
				   KODI */
    Headers,
    Dump,
    Parameters,
    Agent,
    AgentAttributes,
    AgentStatus,
    AgentProgram,
    Actions,
    MapComplete,
    MapIncremental,
    MapAgents,
    MapFood,
    MapCompleteXPM,
    MapAgentsXPM,
    MapFoodXPM,
    Statistics,
    Times,
    BirthAndDeath,
    FilePositions
  };


char *DumpParameters[] =
{
  "TotalDumpAtEnd",
  "DisplayHeaders", "DisplayDump", "DisplayParameters",
  "DisplayAgents",
  "DisplayAgentAttributes", "DisplayAgentStatus", "DisplayAgentProgram",
  "DisplayActions",
  "DisplayMapComplete", "DisplayMapIncremental",
  "DisplayMapAgents", "DisplayMapFood",
  "DisplayMapCompleteXPM", "DisplayMapAgentsXPM", "DisplayMapFoodXPM",
  "DisplayStatistics", "DisplayTimes", "DisplayBirthAndDeath",
  "FilePositions"
};

static unsigned NumberOfCalls[DISPLAY_PARAMETERS];

static CountType FirstTimeStep = -1, CurrentTimeStep = -1;
static int DisplayNow[DISPLAY_PARAMETERS];


DumpChannel *
Dump_FindChannel (char *Name)
{
  DumpChannel *channel;

  for (channel = Channels; channel; channel = channel->next)
    if (strcmp (channel->name, Name) == 0)
      break;

  return channel;
}

DumpChannel *
Dump_CreateChannel (char *Name, IO_ChannelDomainType Type, char *Mode, int Flush)
{
  DumpChannel *new_channel;

  if ((new_channel = Dump_FindChannel (Name)))
    return new_channel;

  if ((new_channel = Allocate (sizeof (DumpChannel))))
    {
      unsigned i;

      strcpy (new_channel->name, Name);
      new_channel->stream = IO_OpenChannel (Type, Name, Mode);
      new_channel->type = Type;
      strcpy (new_channel->mode, Mode ? Mode : "w");
      new_channel->flush_immediately = Flush;
      switch (Flush)
	{
	case _IONBF:
	  setbuf (new_channel->stream, NULL);
	  break;
	case _IOLBF:
	  break;
	case _IOFBF:
	default:
	  break;
	}
      new_channel->filePosStream = NULL;

      for (i = 0; i < DISPLAY_PARAMETERS; i++)
	new_channel->display[i] = Default_Channel.display[i];

      new_channel->next = Channels;
      Channels = new_channel;
    }

  return new_channel;
}

void 
Dump_SetChannelHeader (DumpChannel * Channel, DynaStringType * Header)
{
  Channel->header = Header;
}

static DumpChannel *
Dump_Close (DumpChannel * Channel)
{
  if (Channel->filePosStream)
    {
      fclose (Channel->filePosStream);
      Channel->filePosStream = NULL;
    }

  if (!IO_CloseChannel (Channel->type, Channel->name))
    Warning ("Dump_CloseChannel", "channel %s could not be closed\n",
	     Channel->name ? Channel->name : "");

  Channel->stream = NULL;

  return Channel;
}

DumpChannel *
Dump_CloseChannel (char *Name)
{
  DumpChannel *channel;

  if ((channel = Dump_FindChannel (Name)))
    return Dump_Close (channel);

  return NULL;
}

DumpChannel *
Dump_ReopenChannel (char *Name, char *Mode)
{
  DumpChannel *channel;

  if ((channel = Dump_FindChannel (Name)))
    {
      if (!channel->stream)
	{
	  if (Mode)
	    strcpy (channel->mode, Mode);

	  channel->stream = IO_OpenChannel (channel->type, Name,
					    channel->mode);
	}
      else
	Warning ("Dump_ReopenChannel",
		 "can't reopen %s 'cause it's still opened\n", Name);

      return channel;
    }

  return NULL;
}


void 
Dump_DestroyChannels ()
{
  DumpChannel *channel;

  for (channel = Channels; channel;)
    {
      Channels = channel->next;
      if (channel->header)
	DynaString_Free (channel->header);
      Free (channel);
      channel = Channels;
    }
}

static DynaStringType *
PrintChannel (DumpChannel * channel, DynaStringType * buffer)
{
  unsigned i;

  if (!buffer)
    buffer = DynaString_Alloc (512);

  DynaString_Append (buffer, "CHANNEL ");
  if (strchr (channel->name, ' '))
    DynaString_AppendFormat (buffer, strlen (channel->name) + 3,
			     "%c%s%c", '"', channel->name, '"');
  else
    DynaString_Append (buffer, channel->name);

  DynaString_Append (buffer, " TYPE ");

  if (channel->type == IO_SocketChannel)
    DynaString_Append (buffer, "SOCKET ");
  else if (channel->type == IO_FileChannel)
    DynaString_Append (buffer, "FILE ");
  else if (channel->type == IO_PipeChannel)
    DynaString_Append (buffer, "PIPE ");
  else if (channel->type == IO_CommandChannel)
    DynaString_Append (buffer, "COMMAND ");

  DynaString_Append (buffer, " MODE ");
  DynaString_Append (buffer, channel->mode[0] == 'a' ? "APPEND " :
		     "OVERWRITE ");

  if (channel->flush_immediately)
    DynaString_Append (buffer, "FLUSH\n");
  else
    DynaString_Append (buffer, "\n");

  for (i = 0; i < DISPLAY_PARAMETERS; i++)
    {
      DynaString_Append (buffer, "\t");
      DynaString_Append (buffer, DumpParameters[i]);
      DynaString_Append (buffer, "\t = ");
      DynaString_AppendInt (buffer, channel->display[i]);
      DynaString_Append (buffer, "\n");
    }

  DynaString_Append (buffer, "END_CHANNEL\n");

  return buffer;
}

static bool 
Dump_PrintOnChannel (va_alist)
     va_dcl
{
  va_list arglist;
  DumpChannel *channel;
  char *format;

  va_start (arglist);

  channel = va_arg (arglist, DumpChannel *);
  format = va_arg (arglist, char *);
  if (vfprintf (channel->stream, format, arglist) == EOF)
    {
      if (errno == EPIPE)
	{
	  Warning ("[Dump_PrintOnChannel]",
		   "failure %d on channel %s ... closing\n",
		   errno, channel->name ? channel->name : "unnamed");
	  Dump_Close (channel);
	}
      else
	{
	  Warning ("[Dump_PrintOnChannel]",
		   "failure %d on channel %s with:\n",
		   errno, channel->name ? channel->name : "unnamed");
	  Warning ("", format, arglist);
	}
      return FALSE;
    }
  else
    return TRUE;
  va_end (arglist);
}

bool 
Dump_SetParameterValueForChannel (DumpChannel * Channel,
				  string Name, string Value)
{
  unsigned i;

  if (!Channel)
    Channel = &Default_Channel;

  for (i = 0; i < DISPLAY_PARAMETERS; i++)
    {
      if (strcmp (DumpParameters[i], Name) == 0)
	{
	  Channel->display[i] = atoi (Value);
	  return TRUE;
	}
    }

  Warning ("Dump", "Unknown Parameter %s\n", Name);

  return FALSE;
}

bool 
Dump_SetParameterValue (string Name, string Value)
{
  int index;

  if (!parameterKeys)
    parameterKeys = StringList_Alloc (50);

  if (!parameterValues)
    parameterValues = StringList_Alloc (50);

  if ((index = StringList_FindString (parameterKeys, Name)) >= 0)
    StringList_ReplaceWithString (parameterValues, strdup (Value), index);
  else
    {
      index = StringList_AddString (parameterKeys, strdup (Name));
      assert (StringList_AddString (parameterValues, strdup (Value)) == index);
    }

  return FALSE;
}

/**************************************************************************
 * Dump_ConcludeTimeStep should dump at the end of the timestep
 *
 *************************************************************************/
void 
Dump_ConcludeTimeStep (CountType TimeStep)
{
  Dump_DumpSection ();
}

void 
Dump_EndSimulation ()
{
  DumpChannel *channel;

  if (simState != End)
    {
      time_t now;

      time (&now);
      strcpy (EndTime, ctime (&now));
      simState = End;
    }

  /* set all DumpVariables to TRUE (for totaldump at simulation end) */
  for (channel = Channels; channel; channel = channel->next)
    {
      unsigned i;

      if (channel->display[TotalDumpAtEnd] != DISPLAY_NEVER)
	{
	  for (i = 0; i < DISPLAY_PARAMETERS; i++)
	    {
	      DisplayNow[i] = 1;
/*                              channel->display[i] = 1; */
	      NumberOfCalls[i] = 0;
	    }
	}
    }

  Dump_DumpSection ();
}

void 
Dump_CloseFiles ()
{
  DumpChannel *channel;

  for (channel = Channels; channel; channel = channel->next)
    {
      if (channel->filePosStream)
	fclose (channel->filePosStream);

      if (channel->stream)
	if (!IO_CloseChannel (channel->type, channel->name))
	  Warning ("Dump_CloseFiles", "File %s could not be closed\n",
		   channel->name);
    }

  Dump_DestroyChannels ();
/*      Dump_DestroyWorldFiles(); */
}

void 
Dump_DumpSection ()
{
  if (Dump_WantDumpHeader ())
    Dump_DumpHeader (NULL);

  if (Dump_WantDumpDump ())
    Dump_DumpDump (NULL);

  if (Dump_WantDumpParameters ())
    Dump_DumpParameters (NULL);
}


bool 
Dump_ChannelOK (DumpChannel * channel)
{
  return channel != NULL;
}

bool 
Dump_ChannelWantDump (unsigned Section, DumpChannel * channel)
{
  if (channel->stream)
    {
      switch (simState)
	{
	case Running:
	  switch (channel->display[Section])
	    {
	    case DISPLAY_UNSET:
	    case DISPLAY_NEVER:
	      return FALSE;
	      break;
	    case DISPLAY_ONCE:
	      if (FirstTimeStep == CurrentTimeStep)
		return TRUE;
	      else
		return FALSE;
	      break;
	    default:
	      if (((CurrentTimeStep - FirstTimeStep) %
		   channel->display[Section]) == 0)
		return TRUE;
	      else
		return FALSE;
	      break;
	    }
	  break;
	case End:
	  return channel->display[TotalDumpAtEnd];
	  break;
	case Startup:
	  /*                      return channel->display[TotalDumpAtStart];
	     break; */
	default:
	  return FALSE;
	  break;
	}
    }

  return FALSE;
}

bool 
Dump_WantDumpHeader ()
{
  return DisplayNow[Headers] != 0;
}

bool 
Dump_WantDumpDump ()
{
  return DisplayNow[Dump] != 0;
}

bool 
Dump_WantDumpParameters ()
{
  return DisplayNow[Parameters] != 0;
}

bool 
Dump_WantDumpAgentsSection ()
{
  return DisplayNow[Agent] != 0;
}

bool 
Dump_WantDumpAgentAttributes ()
{
  return DisplayNow[AgentAttributes] != 0;
}

bool 
Dump_WantDumpAgentStatus ()
{
  return DisplayNow[AgentStatus] != 0;
}

bool 
Dump_WantDumpAgentProgram ()
{
  return DisplayNow[AgentProgram] != 0;
}

bool 
Dump_WantDumpActions ()
{
  return DisplayNow[Actions] != 0;
}

bool 
Dump_WantDumpMapComplete ()
{
  return DisplayNow[MapComplete] != 0;
}

bool 
Dump_WantDumpMapIncremental ()
{
  return DisplayNow[MapIncremental] != 0;
}

bool 
Dump_WantDumpMapAgents ()
{
  return DisplayNow[MapAgents] != 0;
}

bool 
Dump_WantDumpMapFood ()
{
  return DisplayNow[MapFood] != 0;
}

bool 
Dump_WantDumpMapCompleteXPM ()
{
  return DisplayNow[MapCompleteXPM] != 0;
}

bool 
Dump_WantDumpMapAgentsXPM ()
{
  return DisplayNow[MapAgentsXPM] != 0;
}

bool 
Dump_WantDumpMapFoodXPM ()
{
  return DisplayNow[MapFoodXPM] != 0;
}

bool 
Dump_WantDumpStatistics ()
{
  return DisplayNow[Statistics] != 0;
}

bool 
Dump_WantDumpTimes ()
{
  return DisplayNow[Times] != 0;
}

bool 
Dump_WantDumpBirthAndDeath ()
{
  return DisplayNow[BirthAndDeath] != 0;
}


bool 
Dump_Dump (unsigned Section, string Buffer)
{
  DumpChannel *channel;

  if (Section < DISPLAY_PARAMETERS)
    {
      for (channel = Channels; channel; channel = channel->next)
	{
	  if (channel->stream && Dump_ChannelWantDump (Section, channel))
	    {
	      if (Dump_ChannelWantDump (FilePositions, channel) &&
		  channel->filePosStream && (NumberOfCalls[Section] == 0))
		{
		  long int file_pos;

		  file_pos = ftell (channel->stream);
		  switch (Section)
		    {
		    case Agent:
		      fprintf (channel->filePosStream,
			       "Agents = %ld\n", file_pos);
		      break;
		    case MapIncremental:
		      fprintf (channel->filePosStream,
			       "MapIncremental = %ld\n", file_pos);
		      break;
		    case MapComplete:
		      fprintf (channel->filePosStream,
			       "MapComplete = %ld\n", file_pos);
		      break;
		    case MapAgents:
		      fprintf (channel->filePosStream,
			       "MapAgents = %ld\n", file_pos);
		      break;
		    case MapFood:
		      fprintf (channel->filePosStream,
			       "MapFood = %ld\n", file_pos);
		      break;
		    case Actions:
		      fprintf (channel->filePosStream,
			       "Actions = %ld\n", file_pos);
		      break;
		    default:
		      break;
		    }
		}

	      DebugInfo ("%d: Dumping %s to channel %s\n", CurrentTimeStep,
			 DumpParameters[Section], channel->name);
	      DebugInfo ("%s", Buffer);
	      Dump_PrintOnChannel (channel, "%s", Buffer);

	      if (channel->flush_immediately)
		fflush (channel->stream);
	    }
	}

      NumberOfCalls[Section]++;
      return TRUE;
    }
  else
    return FALSE;
}

bool 
Dump_DumpDump (string Buffer)
{
  DumpChannel *channel;
  DynaStringType *channels;

  channels = DynaString_Alloc (1024);

  DynaString_Append (channels, "DUMP\n");
  for (channel = Channels; channel; channel = channel->next)
    PrintChannel (channel, channels);
  DynaString_Append (channels, "END_DUMP\n");

  Dump_Dump (Dump, DynaString_CString (channels));
  DynaString_Free (channels);

  return TRUE;
}

bool 
Dump_DumpParameters (string Buffer)
{
  DynaStringType *parameters;
  unsigned index;
  string key, value;

  parameters = DynaString_Alloc (1024);
  DynaString_Append (parameters, "PARAMETERS\n");
  for (index = 0; TRUE; index++)
    {
      if ((key = StringList_StringAt (parameterKeys, index)))
	{
	  value = StringList_StringAt (parameterValues, index);

	  DynaString_Append (parameters, "\t");
	  DynaString_Append (parameters, key);
	  DynaString_Append (parameters, "\t= ");
	  DynaString_Append (parameters, value);
	  DynaString_Append (parameters, "\n");
	}
      else
	break;
    }
  DynaString_Append (parameters, "END_PARAMETERS\n");

  Dump_Dump (Parameters, DynaString_CString (parameters));

  Free (DynaString_CStringAndFree (parameters));

  return TRUE;
}

bool 
Dump_DumpAgent (string Buffer)
{
  return Dump_Dump (Agent, Buffer);
}

bool 
Dump_DumpAgentAttributes (string Buffer)
{
  return Dump_Dump (AgentAttributes, Buffer);
}

bool 
Dump_DumpAgentStatus (string Buffer)
{
  return Dump_Dump (AgentStatus, Buffer);
}

bool 
Dump_DumpAgentProgram (string Buffer)
{
  return Dump_Dump (AgentProgram, Buffer);
}

bool 
Dump_DumpActions (string Buffer)
{
  return Dump_Dump (Actions, Buffer);
}

bool 
Dump_DumpMapComplete (string Buffer)
{
  return Dump_Dump (MapComplete, Buffer);
}

bool 
Dump_DumpMapIncremental (string Buffer)
{
  return Dump_Dump (MapIncremental, Buffer);
}

bool 
Dump_DumpMapAgents (string Buffer)
{
  return Dump_Dump (MapAgents, Buffer);
}

bool 
Dump_DumpMapFood (string Buffer)
{
  return Dump_Dump (MapFood, Buffer);
}

bool 
Dump_DumpMapCompleteXPM (string Buffer)
{
  return Dump_Dump (MapCompleteXPM, Buffer);
}

bool 
Dump_DumpMapAgentsXPM (string Buffer)
{
  return Dump_Dump (MapAgentsXPM, Buffer);
}

bool 
Dump_DumpMapFoodXPM (string Buffer)
{
  return Dump_Dump (MapFoodXPM, Buffer);
}


bool 
Dump_DumpStatistics (string Buffer)
{
  return Dump_Dump (Statistics, Buffer);
}

bool 
Dump_DumpTimes (string Buffer)
{
  return Dump_Dump (Times, Buffer);
}

bool 
Dump_DumpBirthAndDeath (string Buffer)
{
  return Dump_Dump (BirthAndDeath, Buffer);
}

static struct timeval 
hertzToTimeval (int hertz)
{
  struct timeval result;

  result.tv_sec = hertz / 60;
  result.tv_usec = hertz % 60;

  return result;
}

static struct timeval 
subtract_times (struct timeval tv1, struct timeval tv2)
{
  struct timeval result;

  if (tv2.tv_usec > tv1.tv_usec)
    {
      result.tv_sec = tv1.tv_sec - tv2.tv_sec - 1;
      result.tv_usec = 1000000 - (tv2.tv_usec - tv1.tv_usec);
    }
  else
    {
      result.tv_sec = tv1.tv_sec - tv2.tv_sec;
      result.tv_usec = tv1.tv_usec - tv2.tv_usec;
    }

  return result;
}

static void 
fprintt (FILE * stream, struct timeval timeval)
{
  unsigned int hours, minutes;

  hours = timeval.tv_sec / 3600;
  timeval.tv_sec -= hours * 3600;

  minutes = timeval.tv_sec / 60;
  timeval.tv_sec -= minutes * 60;

  fprintf (stream, "%d:%d:%ld:%ld", hours, minutes,
	   timeval.tv_sec, timeval.tv_usec / 100);
}

static void 
sprintt (char *buffer, struct timeval timeval)
{
  unsigned int hours, minutes;

  hours = timeval.tv_sec / 3600;
  timeval.tv_sec -= hours * 3600;

  minutes = timeval.tv_sec / 60;
  timeval.tv_sec -= minutes * 60;

  sprintf (buffer, "%d:%d:%ld:%ld", hours, minutes,
	   timeval.tv_sec, timeval.tv_usec);
}

bool 
Dump_DumpHeader (string Buffer)
{
  DumpChannel *channel;
  DynaStringType *common_header;

  Dump_Dump (Headers, "HEADER\n");

/*      dump common header
 *    includes the time the simulation started and ended
 *      and more...?
 */

  common_header = DynaString_Alloc (512);

  DynaString_Append (common_header, "Philia V0.9\n");
  DynaString_Append (common_header, "DumpFormat V0.9\n\n");

  if (simState != Startup)
    {
      DynaString_Append (common_header, "Started at:\t");
      DynaString_Append (common_header, StartTime);
    }

  if (simState == End)
    {
      DynaString_Append (common_header, "Ended at:\t");
      DynaString_Append (common_header, EndTime);
    }

  Dump_Dump (Headers, DynaString_CString (common_header));
  DynaString_Free (common_header);

  Dump_Dump (Headers, "\n\n");

/*      dump channel specific header
 *    i.e. the header that was in the worldfile that defined
 *      the channel
 *      and the definition of the channel
 */
  for (channel = Channels; channel; channel = channel->next)
    {
      DynaStringType *channel_header;

      if (Dump_ChannelWantDump (Headers, channel))
	{
	  fputs ("\nThis dumpfile was generated according to:\n",
		 channel->stream);
	  channel_header = PrintChannel (channel, NULL);
	  Dump_PrintOnChannel (channel, "%s", DynaString_CString (channel_header));
	  DynaString_Free (channel_header);

	  Message ("%d: Dumping %s to channel %s\n", CurrentTimeStep,
		   DumpParameters[Headers], channel->name);

	  if (channel->header != NULL)
	    {
	      Dump_PrintOnChannel (channel, "%s",
				   DynaString_CString (channel->header));
	    }

	}
    }

  Dump_Dump (Headers, "END_HEADER\n");

  return TRUE;
}

void 
Dump_InitTimeStep (CountType TimeStep)
{
  bool dump_statistics, dump_times;

  DumpChannel *channel;
  unsigned i;
  int n;

  CurrentUserTime = Res_GetUserTime ();
  CurrentMemory = Res_GetMemoryUsage () / 1024;

  if (simState != Running)
    {
      time_t now;

      time (&now);
      strcpy (StartTime, ctime (&now));

      StartUserTime = LastUserTime = CurrentUserTime;
      StartMemory = LastMemory = CurrentMemory;

      simState = Running;
    }


  if (FirstTimeStep == -1)
    FirstTimeStep = TimeStep;
  CurrentTimeStep = TimeStep;

  for (i = 0; i < DISPLAY_PARAMETERS; i++)
    {
      DisplayNow[i] = 0;
      NumberOfCalls[i] = 0;
    }

  for (channel = Channels; channel; channel = channel->next)
    {
      for (i = 0, n = 0; i < DISPLAY_PARAMETERS; i++)
	if (Dump_ChannelWantDump (i, channel))
	  {
	    n++;
	    DisplayNow[i] = 1;
	  }

      if ((n - (channel->display[TotalDumpAtEnd] != 0 ? 1 : 0)) > 0)
	{
	  char filename[MAXPATHLEN + 1];

	  if (Dump_ChannelWantDump (FilePositions, channel))
	    {
	      if (!channel->filePosStream)
		{
		  strcpy (filename, channel->name);
		  strcat (filename, ".fp");

		  channel->filePosStream = fopen (filename, channel->mode);
		}

	      if (channel->filePosStream)
		fprintf (channel->filePosStream, "Timestep %d = %ld\n",
			 TimeStep, ftell (channel->stream));
	      else
		Warning ("Dump_InitTimestep",
			 "can't open file %s for dumping filepositions "
			 "of channel %s\n", filename, channel->name);
	    }

	  dump_statistics = Dump_ChannelWantDump (Statistics, channel);
	  dump_times = Dump_ChannelWantDump (Times, channel);

	  if (dump_statistics || dump_times)
	    {
	      Dump_PrintOnChannel (channel, "#! Timestep = %d\n", TimeStep);
	    }

	  if (Dump_ChannelWantDump (Times, channel))
	    {
	      char current_time[32];
	      struct rusage rubuf;

	      getrusage (RUSAGE_SELF, &rubuf);
	      Dump_PrintOnChannel (channel, "#! PageFaults = %ld\n",
				   rubuf.ru_majflt);
	      Res_TimevalToString (Res_SubtractTimes (CurrentUserTime,
					       LastUserTime), current_time);
	      Dump_PrintOnChannel (channel, "#! TimeLast = %s\n",
				   current_time);

	      Res_TimevalToString (Res_SubtractTimes (CurrentUserTime,
					      StartUserTime), current_time);
	      Dump_PrintOnChannel (channel, "#! TimeTotal = %s\n",
				   current_time);

	      Dump_PrintOnChannel (channel, "#! MemoryLast = %dK\n",
				   CurrentMemory - LastMemory);
	      Dump_PrintOnChannel (channel, "#! MemoryTotal = %dK\n",
				   CurrentMemory);

	      LastUserTime = CurrentUserTime;
	      LastMemory = CurrentMemory;
	    }

	  if (channel->flush_immediately)
	    fflush (channel->stream);
	}
    }

  DisplayNow[Agent] = DisplayNow[Agent] || DisplayNow[AgentStatus] ||
    DisplayNow[AgentAttributes] || DisplayNow[AgentProgram];

  for (i = 0; i < DISPLAY_PARAMETERS; i++)
    DebugInfo ("%d: want dump of %s\n", CurrentTimeStep, DumpParameters[i]);

}
