/*-module-*****************************************************************
 *
 * Modul     : Interpreter (FUNKTIONSMODUL)
 *
 * Beschreib.:
 *
 * Autor     :erik
 *
 * Datum     : 8.1.1996
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"

#include "../agents/AgentTypeDefs.h"
#include "../agents/AgentSet.h"
#include "../program/AgentPrgFunc.h"
#include "../dump/Dump.h"
#include "../random/Random.h"

#include "Interpreter.h"
#include "InterpreterPrivate.h"
#include "FuncSem.h"

/*************************************************************************/

#define YES 1
#define NO  0
#define STR_EQUAL 0

#define RAND() ((int) Random_Rand() & 0xffff)

/*************************************************************************/

DynaStringType *ActionDumpBuffer = NULL;

/*************************************************************************/

typedef int
  PrgResourceType;


typedef struct FuncSpecType
  {
    const string name;
    short arity;
    PrgResourceType resources;
      bool (*semantics) (AgentIdType, NodePtrArrayType);
  }
FuncSpecType;


/*************************************************************************/

#define NUM_FUNCTIONS		28
#define NUM_LEAF_FUNCTIONS	22

static PrgResourceType AgentInitialResources = 0;

static FuncSpecType FunctionTable[NUM_FUNCTIONS] =
{
  {"Nop", 0, 1, FuncSem_Nop},

  {"Seq", 3, 1, FuncSem_Seq},
  {"If", 3, 1, FuncSem_If},

  {"And", 2, 1, FuncSem_And},
  {"Or", 2, 1, FuncSem_Or},
  {"Not", 1, 1, FuncSem_Not},

  {"Attack", 0, 6, FuncSem_Attack},
  {"Mate", 0, 6, FuncSem_Mate},
  {"Eat", 0, 2, FuncSem_Eat},
  {"MakeFriend", 0, 2, FuncSem_MakeFriend},
  {"Scream", 0, 2, FuncSem_Scream},
  {"Move", 0, 6, FuncSem_Move},
  {"TurnLeft", 0, 2, FuncSem_TurnLeft},
  {"TurnRight", 0, 2, FuncSem_TurnRight},
  {"TurnSound", 0, 2, FuncSem_TurnSound},
  {"SetMem", 1, 2, FuncSem_SetMem},

  {"Hungry?", 0, 1, FuncSem_Hungry},
  {"UnderAttack?", 0, 1, FuncSem_UnderAttack},
  {"Sound?", 0, 1, FuncSem_Sound},
  {"Mem?", 0, 1, FuncSem_Mem},
  {"AgentHere?", 0, 1, FuncSem_AgentHere},
  {"AgentAhead?", 0, 1, FuncSem_AgentAhead},
  {"FoodHere?", 0, 1, FuncSem_FoodHere},
  {"FoodAhead?", 0, 1, FuncSem_FoodAhead},
  {"FriendHere?", 0, 1, FuncSem_FriendHere},
  {"FriendAhead?", 0, 1, FuncSem_FriendAhead},
  {"NothingHere?", 0, 1, FuncSem_NothingHere},
  {"NothingAhead?", 0, 1, FuncSem_NothingAhead}};


static int SimpleFuncs[] =
{
  0, 6, 7, 8, 9,
  10, 11, 12, 13, 14,
  16, 17, 18, 19, 20,
  21, 22, 23, 24, 25,
  26, 27
};

static int ComplexFuncs[] =
{
  1, 2, 3, 4, 5, 15
};

static double vmutation = 0.0;
static short vmute = 0;
static double rmutation = 0.0;
static short rmute = 0;
static PrgResourceType AgentPrgResources;
static AgentIdType ProperAgentId;

static unsigned MinProgramDepth = 2, MaxProgramDepth = 5;
static Mutation = 0;

/*-ftn-********************************************************************
 *
 * Interpreter_FunctionNamed (PRIVAT)
 *
 * Liefert die Funktion mit dem uebergebenen Namen.
 *
 * #fertig
 *
 *************************************************************************/

CommandType 
Interpreter_FunctionNamed (const string Name)
{
  CommandType funcID;

  for (funcID = 0; funcID < NUM_FUNCTIONS; funcID++)
    if (strcmp (FunctionTable[funcID].name, Name) == STR_EQUAL)
      break;
  return (funcID == NUM_FUNCTIONS) ? -1 : funcID;
}



/*-ftn-********************************************************************
 *
 * Interpreter_*OfFunction (PRIVAT)
 *
 * Liefert den Namen bzw. die Aritaet der uebergebenen Funktion
 *
 * #fertig
 *
 *************************************************************************/

const string 
Interpreter_NameOfFunction (CommandType Func)
{
  return (Func >= 0 && Func < NUM_FUNCTIONS) ? FunctionTable[Func].name : NULL;
}



int 
Interpreter_ArityOfFunction (CommandType Func)
{
  return (Func >= 0 && Func < NUM_FUNCTIONS) ? FunctionTable[Func].arity : -1;
}


/*-ftn-********************************************************************
 *
 * Interpreter_SetParameterValue
 *
 * Setzt die fuer das Interpreter-Modul relevanten World Parameter. Zur
 * Zeit nur `MaxCost,' das angibt wieviel Resourcen fuer die Abarbeitung
 * eines Programms den Agenten in jedem Zeitschritt zur Verfuegung stehen.
 * Ist dieser Wert kleiner oder gleich 0 wird ein Fehler generiert.
 *
 * Auserdem wird mit Parametern der Form `Cost-*' der Ressourcenbedarf
 * der einzelnen Befehle gesetzt.
 *
 * #fertig
 *
 *************************************************************************/

bool 
Interpreter_SetParameterValue (string Key, string Value)
{
  PrgResourceType tmpVal;
  bool wantedVal = NO;
  CommandType cmd;

  if (strcmp (Key, "MaxCost") == STR_EQUAL)
    {
      tmpVal = atoi (Value);
      if (tmpVal <= 0)
	Error ("Interpreter_SetParameterValue",
	       "World parameter 'MaxCost' must be greater than 0.\n");
      AgentInitialResources = tmpVal;
      wantedVal = YES;
    }
  else if (strcmp (Key, "VMutation") == STR_EQUAL)
    {
      if ((vmutation = atof (Value)) < 0)
	Error ("Interpreter_SetParameterValue", "VMutation must be greater than 0.\n");
      vmute = 1;
      wantedVal = YES;
    }
  else if (strcmp (Key, "RMutation") == STR_EQUAL)
    {
      if ((rmutation = atof (Value)) < 0)
	Error ("Interpreter_SetParameterValue", "RMutation must be greater than 0.\n");
      rmute = 1;
      wantedVal = YES;
    }
  else if (strncmp (Key, "Cost-", 5) == STR_EQUAL)
    {
      cmd = Interpreter_FunctionNamed (Key + 5);
      if (cmd < 0)
	Error ("Interpreter_SetParameterValue",
	       "Specified cost for unknown command %s.\n", Key + 5);
      tmpVal = atoi (Value);
      if (tmpVal < 0)
	Error ("Interpreter_SetParameterValue",
	       "Function cost cannot be negative.\n");
      FunctionTable[cmd].resources = tmpVal;
      wantedVal = YES;
    }
  else if (strcmp (Key, "MinProgramDepth") == 0)
    {
      MinProgramDepth = atoi (Value);
      wantedVal = YES;
    }
  else if (strcmp (Key, "MaxProgramDepth") == 0)
    {
      MaxProgramDepth = atoi (Value);
      wantedVal = YES;
    }
  else if (strcmp (Key, "Mutation") == 0)
    {
      Mutation = atoi (Value);
      wantedVal = YES;
    }

  return wantedVal;
}



/*-ftn-********************************************************************
 *
 * Interpreter_InitTimestep
 *
 * Bereitet den Puffer fuer das Action Dump vor. (Falls gewuenscht.)
 *
 * #fertig
 *
 *************************************************************************/

void 
Interpreter_InitTimestep (CountType Timestep)
{
  if (Dump_WantDumpActions ())
    {
      if (ActionDumpBuffer != NULL)
	DynaString_Reset (ActionDumpBuffer);
      else
	ActionDumpBuffer = DynaString_Alloc (4096);
      DynaString_Append (ActionDumpBuffer, "ACTIONS\n");
    }
  else
    {
      if (ActionDumpBuffer != NULL)
	{
	  DynaString_Free (ActionDumpBuffer);
	  ActionDumpBuffer = NULL;
	}
    }
}


/*-ftn-********************************************************************
 *
 * Interpreter_ConcludeTimestep
 *
 * Gibt das waehrend des Zeitschritts aufgesammelte Action Dump aus.
 *
 * #fertig
 *
 *************************************************************************/

void 
Interpreter_ConcludeTimestep (CountType Timestep)
{
  if (ActionDumpBuffer != NULL)
    {
      DynaString_Append (ActionDumpBuffer, "END_ACTIONS\n");
      Dump_DumpActions (DynaString_CString (ActionDumpBuffer));
    }
}


/*-ftn-********************************************************************
 *
 * Interpreter_EvalAgent
 *
 * Fuehrt das Programm des uebergebenen Agenten aus. Dazu wird zunaechst
 * das Programm des Agenten geholt, dann der Zeiger auf den Befehl an der 
 * Wurzel des Programmbaumes ermittelt und damit dann die allgemeine 
 * Auswertungsroutine aufgerufen.
 *
 * #ungetestet
 *
 *************************************************************************/

void 
Interpreter_EvalAgent (AgentIdType AgentId)
{
  ProgramType *program;
  NodePtrType rootNode;
  CountType AgentLastMate;
  CountType Idle;		/* added by Martin */
  
  if (AgentSet_GetEnergy (AgentId) <= (EnergyType) AgentSet_GetMaxEnergy (AgentId) * 
  	AgentSet_GetMinEnergyFactor (AgentId) || AgentSet_GetAttacks (AgentId) == -1)
  	return;
  	/* Verstorbene Agenten besser nicht ausfhren */

  AgentSet_SetScreams (AgentId, 0);
  Idle = AgentSet_GetIdle (AgentId);
  if (Idle > 0)
    {
      AgentSet_SetIdle (AgentId, Idle - 1);
      return;
    }				/* berspringe Agent falls noch immer Pause */

  ProperAgentId = AgentId;
  AgentPrgResources = AgentInitialResources;
  program = AgentSet_GetProgram (AgentId);

  rootNode = AgentPrgFunc_Program2NodePtr (program);
  if (ActionDumpBuffer == NULL)
    {
      Interpreter_Eval (AgentId, rootNode);
    }
  else
    {
      DynaString_AppendFormat (ActionDumpBuffer, 100, "\tAGENT %d\n", AgentId);
      DynaString_AppendFormat (ActionDumpBuffer, 100, "\tTYPE %c\n", AgentSet_GetType (AgentId));
      Interpreter_Eval (AgentId, rootNode);
      DynaString_Append (ActionDumpBuffer, "\tEND_AGENT\n");
    }
    if (AgentSet_GetAttacks (AgentId) != -1)
      AgentSet_SetAttacks (AgentId, 0);

  /* added by Martin */
  AgentLastMate = AgentSet_GetLastMate (AgentId);
  if (AgentLastMate > 0)
    AgentSet_SetLastMate (AgentId, AgentLastMate - 1);
}



/*-ftn-********************************************************************
 *
 * Interpreter_Eval
 *
 * Fuehrt den Befehl an dem uebergebenen Knoten aus. Es wird sichergestellt,
 * dass die uebergebene AgentId gueltig ist, also derjenigen entspricht, mit
 * der Interpreter_EvalAgent zuletzt aufgerufen wurde. Die Funktionen 
 * ist auch fuer die Abrechnung der Kosten (Ressourcen) fuer den Befehl
 * verantwortlich und bricht die Ausfuehrung ab, wenn der Agent nicht mehr
 * genuegend Ressourcen hat, um den Befehl auszufuehren.
 *
 * #ungetestet
 *
 *************************************************************************/

bool 
Interpreter_Eval (AgentIdType AgentId, NodePtrType Node)
{
  CommandType command;
  FuncSpecType *funcSpecPtr;
  NodePtrArrayType args;
  int arity;
  bool retVal;
  ProgramType *program;
  NodePtrType rootNode, newNode;
 
  assert (FUNC_MAX_ARITY == 3);
  assert (AgentId == ProperAgentId);

  if (Node == NULL)
    return NO;

  if (rmute && (Random_DRand () < rmutation))
    {
      program = AgentSet_GetProgram (AgentId);
      rootNode = AgentPrgFunc_Program2NodePtr (program);
      if (rootNode != Node)
	{
	  newNode = Interpreter_CreateRandomNode (0, 0);
	  AgentPrgFunc_Replace (rootNode, Node, newNode);
	  AgentPrgFunc_SetProgramLength (program,
					 AgentPrgFunc_NodeCount (rootNode));
	  Node = newNode;
	  if (ActionDumpBuffer != NULL)
	    DynaString_Append (ActionDumpBuffer,
			       "#!\t\tRMutation happened\n");
	}
    }

  if (!vmute || (Random_DRand () > vmutation))
    command = AgentPrgFunc_GetCommand (Node);
  else
    {
      command = Interpreter_RandomCommand (IRFM_LEAF);
      if (ActionDumpBuffer != NULL)
	DynaString_Append (ActionDumpBuffer,
			   "#!\t\tVMutation happened\n");
    }

  funcSpecPtr = &FunctionTable[command];
  AgentPrgResources -= funcSpecPtr->resources;

  if (AgentPrgResources < 0)
    {
      if (ActionDumpBuffer != NULL)
	DynaString_Append (ActionDumpBuffer,
			   "#!\t\tProgramResources = exceeded\n");
      return NO;
    }

  arity = funcSpecPtr->arity;
  args[0] = (arity > 0) ? AgentPrgFunc_GetChild (Node, 0) : NULL;
  args[1] = (arity > 1) ? AgentPrgFunc_GetChild (Node, 1) : NULL;
  args[2] = (arity > 2) ? AgentPrgFunc_GetChild (Node, 2) : NULL;

  if (ActionDumpBuffer == NULL)
    {
      retVal = (*(funcSpecPtr->semantics)) (AgentId, args);
    }
  else if (arity == 0)
    {
      retVal = (*(funcSpecPtr->semantics)) (AgentId, args);
      DynaString_Append (ActionDumpBuffer, "\t\t");
      DynaString_Append (ActionDumpBuffer, Interpreter_NameOfFunction (command));
      DynaString_Append (ActionDumpBuffer, retVal ? " 1\n" : " 0\n");
    }
  else
    {
      string funcName;

      funcName = Interpreter_NameOfFunction (command);
      DynaString_Append (ActionDumpBuffer, "\t\t");
      DynaString_Append (ActionDumpBuffer, funcName);
      DynaString_Append (ActionDumpBuffer, "\n");
      retVal = (*(funcSpecPtr->semantics)) (AgentId, args);
      DynaString_Append (ActionDumpBuffer, "\t\tEnd");
      DynaString_Append (ActionDumpBuffer, funcName);
      DynaString_Append (ActionDumpBuffer, retVal ? " 1\n" : " 0\n");
    }

  return retVal;
}


/*-ftn-********************************************************************
 *
 * Interpreter_Random*
 *
 * Funktionen zur Erzeugung von zufaelligen Programmen.
 *
 * #ungetestet
 *
 *************************************************************************/

CommandType 
Interpreter_RandomCommand (int what)
{
  int funcId;

  switch (what)
    {
    case 1:
      funcId = ComplexFuncs[RAND () % (NUM_FUNCTIONS - NUM_LEAF_FUNCTIONS)];
      break;
    case 2:
      funcId = SimpleFuncs[RAND () % (NUM_LEAF_FUNCTIONS)];
      break;
    case 3:
      funcId = RAND () % NUM_FUNCTIONS;
      break;
    default:
      funcId = 0;
      break;
    }

  return funcId;
}


NodePtrType 
Interpreter_CreateRandomNode (int MinDepth, int MaxDepth)
{
  int i;
  NodePtrType node;
  CommandType cmd;

  if (MaxDepth >= 0)
    {
      if (MinDepth > 0)
	cmd = Interpreter_RandomCommand (IRFM_INNER);
      else if (MaxDepth > 0)
	cmd = Interpreter_RandomCommand (IRFM_INNER | IRFM_LEAF);
      else
	cmd = Interpreter_RandomCommand (IRFM_LEAF);

      node = AgentPrgFunc_Create (FunctionTable[cmd].arity);
      AgentPrgFunc_Set (node, cmd);
      for (i = 0; i < FunctionTable[cmd].arity; i++)
	{
	  NodePtrType child;

	  child = Interpreter_CreateRandomNode (MinDepth - 1, MaxDepth - 1);
	  AgentPrgFunc_Insert (node, i, child);
	}

      return node;
    }
  else
    Warning ("Interpreter_CreateRandomNode", "MaxDepth = %d\n", MaxDepth);
  return NULL;
}

ProgramType *
Interpreter_CreateRandomProgram ()
{
  CountType length;
  NodePtrType root;

  root = Interpreter_CreateRandomNode (MinProgramDepth, MaxProgramDepth);
  length = AgentPrgFunc_NodeCount (root);
  return AgentPrgFunc_NodePtr2Program (root, length);
}
