
/*-module-*****************************************************************
 *
 * Modul     : Interpreter (FUNKTIONSMODUL)
 *
 * Beschreib.: Agentenprogramm Parser-Submodul. 
 *
 * Autor     : erik
 *
 * Datum     : 26.1.1996
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/
 
%{

#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/DynaString.h"
#include "../misc/Messages.h"
#include "../program/AgentPrgFunc.h"

#include "InterpreterPrivate.h"

static void UnparseNode (NodePtrType Node, DynaStringType *Buffer);

static NodePtrType 		LastNode = NULL;


%}

/*-ftn-********************************************************************
 *
 * Agenten Programm Terminal- und Knoten-Typen
 * 
 * In der Grammatik kommen zwei innere Knotentypen, Funktionsaufrufe
 * und Argumentlisten, sowie drei verschiedene Terminale, Funktionsnamen, 
 * die oeffnende Klammer und die schliessende Klammer, vor.
 *
 * #fertig
 *
 *************************************************************************/

%union
	{
	ArgListType	*yArgListType;
	NodePtrType	yNodePtrType;
	string		yString;
	}
	

%token					EO_PROGRAM
%token 	<yString>		FUNC
%type	<yArgListType>	arglist
%type	<yNodePtrType>	fcall

%%

/*-ftn-********************************************************************
 *
 * Agenten Programm Grammtik
 *
 * Ein gueltiger Funktionsaufruf (fcall) besteht aus dem Funktionsnamen
 * und einer Argumentliste, die von zwei Klammern umschlossen sind. Die
 * Argumentliste (arglist) besteht aus 0 bis FUNC_MAX_ARITY Argumenten
 * vom Typ Funktionsaufruf.
 *
 * Ein Programm ist nur ein Funktionsaufruf, auf den eine Endmarke folgt.
 *
 * #fertig
 *
 *************************************************************************/

program : fcall EO_PROGRAM
			{
			YYACCEPT
			}


fcall : '(' FUNC arglist ')'		
			{		
			$$ = Interpreter_MakeNode($2, $3);
			free($2);
			free($3);
			}

		| FUNC	
			{
			ArgListType	emptyList; 
			
			emptyList.argc = 0;
			$$ = Interpreter_MakeNode($1, &emptyList);
			free($1);
			}		
		;		
 		

arglist :  arglist fcall
			{ 
			$$=$1;
			if($$->argc >= FUNC_MAX_ARITY)
				yyerror("Too many arguments");
			$$->argv[$$->argc] = $2;
			$$->argc++;
			}

		| /* epsilon */
			{
			$$ = calloc(sizeof(ArgListType),1);
			$$->argc = 0;
			}		
		;


%%

/*-ftn-********************************************************************
 *
 * Interpreter_ParseProgram
 *
 * Parst das Agentenprogramm auf dem Stream und baut mit Hilfe der
 * Agent-Program-Functions die interne Repraesentation auf.
 *
 * #fertig
 *
 *************************************************************************/

ProgramType* Interpreter_ParseProgram (file Stream)
	{
	extern file  Interpreter_in;
	CountType    prgLength;
	ProgramType  *program;
	
	LastNode = NULL;
	Interpreter_in = Stream;
	Interpreter_LineNo = 1;
	yyparse();
	prgLength = AgentPrgFunc_NodeCount(LastNode);
	program = AgentPrgFunc_NodePtr2Program(LastNode, prgLength);

	return program;
	}



/*-ftn-********************************************************************
 *
 * Interpreter_MakeNode (PRIVAT)
 *
 * Erzeugt aus einigen Angaben einen neuen Knoten im Programmbaum.
 *
 * #fertig
 *
 *************************************************************************/

NodePtrType Interpreter_MakeNode(const string Name, const ArgListType *P)
	{
	NodePtrType new;
	CommandType	func;
	int			i, arity;
			
	if((func = Interpreter_FunctionNamed(Name)) == -1)
		{
		Interpreter_error("Unknown function: found '%s'", Name);
		func = Interpreter_FunctionNamed("Nop");			
		}	  
	if((arity = Interpreter_ArityOfFunction(func)) != P->argc)
		{
		Interpreter_error("Wrong number of arguments for %s/%d: found %d",
			Interpreter_NameOfFunction(func), arity, P->argc);
		}
		
	new = AgentPrgFunc_Create(P->argc);
	AgentPrgFunc_Set(new, func);
	for(i=0; i<P->argc; i++)
		AgentPrgFunc_Insert(new, i, P->argv[i]);
	LastNode = new;

	return new;	
	}


/*-ftn-********************************************************************
 *
 * Interpreter_error (PRIVAT)
 *
 * Gibt eine vom Parser generierte Fehlermeldung aus.
 *
 * #fertig
 *
 *************************************************************************/

void Interpreter_error(string msgfmt, ...)
	{
	va_list 		args;

	va_start(args, msgfmt);
	fprintf(stderr,"[Program Parser:%d] ", Interpreter_LineNo);
	vfprintf(stderr, msgfmt, args);
	fprintf(stderr,"\n"); 
	}



/*-ftn-********************************************************************
 *
 * Interpreter_UnparseProgram
 *
 * Schreibt die Textversion des Agentenprogramm in einen String.
 * Dieser muss von Aufrufer freigegeben werden.
 *
 * #fertig
 *
 *************************************************************************/

string Interpreter_UnparseProgram (ProgramType *Program)
	{
	DynaStringType 	*buffer;
	
	if(!Program)
		return NULL;

	buffer = DynaString_Alloc(1024);
	DynaString_Append(buffer, "\tPROGRAM\n\t\t");
	Interpreter_UnparseProgramInto(Program, buffer);
	DynaString_Append(buffer, "\n\t\tEND_PROGRAM\n");

	return DynaString_CStringAndFree(buffer);
	}


void Interpreter_UnparseProgramInto (ProgramType *Program, DynaStringType *Buffer)
	{
	NodePtrType     node;

	if((node = AgentPrgFunc_Program2NodePtr(Program)) == NULL)
		DynaString_Append(Buffer, "<empty program>");
	else
		UnparseNode(node, Buffer);
	}

	
	
static void UnparseNode (NodePtrType Node, DynaStringType *Buffer)	
	{
	string 		func;
	NodePtrType	child;
	int			i;
	
	DynaString_Append(Buffer, "(");
	func = Interpreter_NameOfFunction(AgentPrgFunc_GetCommand(Node));
	DynaString_Append(Buffer, (func) ? func : "***");
	for(i=0; i<FUNC_MAX_ARITY; i++)
		if((child = AgentPrgFunc_GetChild(Node, i)) != NULL)
			{
			DynaString_Append(Buffer, " ");
			UnparseNode(child, Buffer);
			}
	DynaString_Append(Buffer, ")");
	}

