/*-module-*****************************************************************
 *
 * Modul     : IO (FUNKTIONSMODUL)
 *
 * Beschreib.:
 *
 * Autor     : erik
 *
 * Datum     : 29.1.1996
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <fcntl.h>

#ifdef linux
#include <arpa/inet.h>
#include <unistd.h>
#endif

#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"
#include "../dump/Dump.h"
#include "IO.h"
#include "../control/Control.h"

/*************************************************************************/

#define YES 1
#define NO	0
#define STR_EQUAL 0

#define PHILIA			"philia"
#define PIPE_PERMISSION	0666

#if HAVE_STRERR
#define IO_ERROR(EXPR, CONTEXT) \
	if(EXPR) { Error("io", "%s (%s)\n", strerror(errno), CONTEXT); }
#else
#define IO_ERROR(EXPR, CONTEXT) \
	if(EXPR) { Error("io", "%s (%s)\n", sys_errlist[errno], CONTEXT); }
#endif

#ifndef MAX
#define MAX(X,Y) (X)>=(Y) ? (X) : (Y);
#endif

/*************************************************************************/

typedef struct ChListNodeType
  {
    IO_ChannelDomainType domain;
    string name;
    file filePtr;
    struct ChListNodeType *next;
  }
ChListNodeType;


/*************************************************************************/

static void SigIOHandler (int signo);
static int AcceptConnectionOnSocket (int Socket, bool allowBlock);
static bool TalkToClient ();
static void SigPipeHandler (int signo);
static int MakePort (int PortNum);


/*************************************************************************/

static volatile bool RequestOnPort = NO;
static int ListenSocket = -1;
static ChListNodeType *OpenChannels = NULL;


/*-ftn-********************************************************************
 *
 * IO_MakePort
 *
 * Diese Funktion initialisiert den TCP Port, auf dem Clients eine
 * Verbindung zu Philia aufnehmen koennen. Ausserdem werden noch 
 * andere Initialisierungen vorgenommen.
 *
 * Liefert TRUE wenn der Port initialisiert werden konnte.
 *
 * #ungetestet
 *
 *************************************************************************/

bool 
IO_MakePort ()
{
  int ret, port;

  /* Socket auf dem well-known Port initialisieren */

  port = Global_UsePort ? Global_UsePort : PHILIA_PORT;
  ListenSocket = MakePort (port);

  /* Asynchrones I/O aktivieren */

  ret = fcntl (ListenSocket, F_SETOWN, getpid ());
  IO_ERROR (ret == -1, "when trying to change to socket options");
#ifdef linux
  ret = fcntl (ListenSocket, F_SETFL, FASYNC);
  IO_ERROR (ret == -1, "when trying to change to socket options");
#endif
/* Hoffentlich ist asynchrone IO bei SUNOS default!? */

  /* Signal Handler installieren */

  RequestOnPort = NO;
  signal (SIGIO, SigIOHandler);
  signal (SIGPIPE, SigPipeHandler);

  return YES;
}


/*-ftn-********************************************************************
 *
 * IO_InitTimestep
 *
 * Behandelt die im letzten Zeitschritt aufgetretenen Verbindungsanfragen
 * von Clients und schliesst automatisch die Kanaele zu inaktiven Clients.
 *
 * #ungetestet
 *
 *************************************************************************/

void 
IO_InitTimestep (CountType TimeStep)
{
  if (RequestOnPort)
    TalkToClient ();
}


/*-ftn-********************************************************************
 *
 * IO_ConcludeTimestep
 *
 * Macht nix.
 *
 * #ungetestet
 *
 *************************************************************************/

void 
IO_ConcludeTimestep (CountType TimeStep)
{
}




/*-ftn-********************************************************************
 *
 * IO_OpenChannel
 *
 * Oeffnet einen neuen Kanal. Der Parameter 'Domain' gibt an, wohin die 
 * Ausgabe erfolgen soll. Moeglich sind zur Zeit 'IO_SocketChannel' und
 * 'IO_FileChannel'. In 'Name' muss bei einer Datei ein gueltiger Filename
 * uebergeben werden. Bei einem Socket die ASCII Darstellung des ent- 
 * sprechenden Ports.
 *	
 * Liefert einen Filedescriptor oder NULL falls der Kanal nicht geoeffnet
 * werden konnte.
 *
 * #ungetestet
 *
 *************************************************************************/

file 
IO_OpenChannel (IO_ChannelDomainType Domain, const string Name,
		const string Mode)
{
  FILE *filePtr = NULL;
  int ret, sock;

  switch (Domain)
    {
    case IO_FileChannel:
      filePtr = fopen (Name, Mode);
      break;
    case IO_SocketChannel:
      sock = MakePort (atoi (Name));
      Message ("Waiting for rendevouz on port '%d'", atoi (Name));
      filePtr = fdopen (AcceptConnectionOnSocket (sock, YES), "w");
      Message ("%s\n", filePtr ? "okay" : "failed");
      break;
    case IO_PipeChannel:
      ret = mknod (Name, S_IFIFO | PIPE_PERMISSION, 0);
      IO_ERROR (ret == -1 && errno != EEXIST, "when creating fifo special file");
      Message ("Waiting for rendevouz on '%s'", Name);
      filePtr = fopen (Name, "w");
      Message ("%s\n", filePtr ? "okay" : "failed");
      break;
    case IO_CommandChannel:
      filePtr = popen (Name, "w");
      break;
    }

  if (filePtr)
    {
      ChListNodeType *new;

      new = malloc (sizeof (ChListNodeType));
      new->domain = Domain;
      new->name = strdup (Name);
      new->filePtr = filePtr;
      new->next = OpenChannels;
      OpenChannels = new;
    }
  return filePtr;
}



/*-ftn-********************************************************************
 *
 * IO_CloseChannel
 *
 * Schliesst einen Kanal. Der Parameter 'Domain' und 'Name' geben an 
 * welcher Kanal geschlossen werden soll.
 *
 * Liefert TRUE wenn der Kanal geschlossen werden konnte, FALSE sonst.
 *
 * #ungetestet
 *
 *************************************************************************/

bool 
IO_CloseChannel (IO_ChannelDomainType Domain, const string Name)
{
  ChListNodeType *nodePtr, *lastNodePtr;

  lastNodePtr = NULL;
  for (nodePtr = OpenChannels; nodePtr != NULL; nodePtr = nodePtr->next)
    {
      if (nodePtr->domain == Domain && strcmp (nodePtr->name, Name) == STR_EQUAL)
	{
	  if (lastNodePtr == NULL)
	    OpenChannels = nodePtr->next;
	  else
	    lastNodePtr->next = nodePtr->next;

	  if (nodePtr->domain == IO_CommandChannel)
	    pclose (nodePtr->filePtr);
	  else
	    {
	      fclose (nodePtr->filePtr);
	      if (nodePtr->domain == IO_PipeChannel)
		unlink (nodePtr->name);
	    }

	  free (nodePtr->name);
	  free (nodePtr);
	  break;
	}
      lastNodePtr = nodePtr;
    }
  return (nodePtr != NULL);
}



/*-ftn-********************************************************************
 *
 * SigIOHandler, AcceptConnectionOnSocket, TalkToClient (PRIVAT)
 *
 * Funktionen, um einen Connection-Request zu erkennen und eine Verbindung
 * aufzubauen. Der Signal-Handler setzt lediglich die modul-globale Variable
 * 'RequestOnPort' auf YES, eine andere Funktion sollte diese Variable
 * regelmaessig ueberpruefen und ggfs. die Funktion 'AnswerConnectionRequests'
 * aufrufen.
 *
 * #ungetestet
 *
 *************************************************************************/

static int 
MakePort (int PortNum)
{
  struct in_addr localAddr;
  struct hostent *localDesc;
  struct sockaddr_in socketAddr;
  int ret, mySocket, one = 1;

  /* Hostbeschreibung ermitteln */

  localAddr.s_addr = inet_addr ("127.0.0.1");
  localDesc = gethostbyaddr ((void *) &localAddr, sizeof (localAddr), AF_INET);
  if (localDesc == NULL)
    return -1;

  /* Socket initialisieren */

  bzero (&socketAddr, sizeof (socketAddr));
  socketAddr.sin_family = localDesc->h_addrtype;
  socketAddr.sin_port = htons (PortNum);
  mySocket = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);
  IO_ERROR (mySocket == -1, "when creating socket");
  ret = setsockopt (mySocket, SOL_SOCKET, SO_REUSEADDR, &one, sizeof (one));
  IO_ERROR (ret == -1, "when setting socket options");

  /* Socket fuer Anfragen vorbereiten */

  ret = bind (mySocket, (struct sockaddr *) &socketAddr, sizeof (socketAddr));
  IO_ERROR (ret == -1, "when binding socket to port");
  ret = listen (mySocket, 5);
  IO_ERROR (ret == -1, "when trying to listing to socket");

  return mySocket;
}


static void 
SigIOHandler (int signo)
{
  RequestOnPort = YES;
}


static int 
AcceptConnectionOnSocket (int Socket, bool allowBlock)
{
  struct sockaddr remoteAddr;
  int remoteAddrSize, oldFlags, ret, newSocket;

  getsockname (Socket, &remoteAddr, &remoteAddrSize);
  if (allowBlock)
    {
      newSocket = accept (Socket, &remoteAddr, &remoteAddrSize);
      IO_ERROR (newSocket < 0, "when waiting for or connecting to client");
    }
  else
    {
      oldFlags = fcntl (Socket, F_GETFL, 0);

#ifdef linux
      ret = fcntl (Socket, F_SETFL, oldFlags | FNDELAY);
#else /* SUNOS */
      ret = fcntl (Socket, F_SETFL, oldFlags | O_NDELAY);
#endif

      IO_ERROR (ret == -1, "when setting socket options for non-blocking accept");
      newSocket = accept (Socket, &remoteAddr, &remoteAddrSize);
      IO_ERROR (newSocket < 0 && errno != EWOULDBLOCK, "when connecting to client");
      ret = fcntl (Socket, F_SETFL, oldFlags);
      IO_ERROR (ret == -1, "when resetting socket options after non-blocking accept");
    }

  return newSocket;
}


static bool 
TalkToClient ()
{
  int channelSocket;

/*      sigblock(sigmask(SIGIO)); */
  while ((channelSocket = AcceptConnectionOnSocket (ListenSocket, NO)) > 0)
    {
      FILE *inFilePtr;
      char buf[100];

      sprintf (buf, "Welcome to Philia!\n", channelSocket);
      write (channelSocket, buf, strlen (buf));
      inFilePtr = fdopen (channelSocket, "r+");
      /* Dump_ParseSection(inFilePtr);  */
      Loader_ParseSection (inFilePtr);
      fclose (inFilePtr);
    }
  RequestOnPort = NO;
/*      sigsetmask(0); */
  return YES;
}



/*-ftn-********************************************************************
 *
 * SigPipeHandler (PRIVAT)
 *
 * #ok
 *
 *************************************************************************/

static void 
SigPipeHandler (int signo)
{
}
